@@ -916,7 +916,7 @@ public class MainFrame extends JFrame {
         miniBrowser.setBounds(50, 100, 700, 400);
         jDesktop.add(miniBrowser);
         
-        TextColor.init();
+        // TextColor.init(size?);  -- should occur when we get font size from prefs 
         
         plotFrame = new PlotFrame(jInputAndPlot);
         
@@ -1197,10 +1197,10 @@ public class MainFrame extends JFrame {
             if (frames[i] instanceof NyquistFile) {
                 NyquistFile nyquistFile = (NyquistFile) frames[i];
                 if (nyquistFile.modified) flag = true;
-                System.out.println("nf " + nyquistFile.getLocation() +
-                                   nyquistFile.getSize() +
-                                   nyquistFile.getFile().getAbsolutePath());
                 if (nyquistFile.getFile() != null) {
+                    System.out.println("nf " + nyquistFile.getLocation() +
+                                       nyquistFile.getSize() +
+                                       nyquistFile.getFile().getAbsolutePath());
                     prefEditFiles = prefEditFiles + "[" +
                         String.valueOf(nyquistFile.getLocation().x) + "," +
                         String.valueOf(nyquistFile.getLocation().y) + "," +
