@@ -81,6 +81,8 @@ public class NyquistFile extends JNonHideableInternalFrame
                     doc.insertString(pos, input, TextColor.attrNormal);
                     pos += input.length();
                 }
+                filePane.snapshot();  // snapshot file so we can redo to
+                // the loaded file after the user undoes back to empty pane
             }
             catch (Exception e2) { System.out.println(e2); }
         }
