@@ -280,7 +280,11 @@ LVAL prepare_audio(LVAL play, SF_INFO *sf_info)
         nyquist_printf("warning: no audio device found\n");
         return NIL;
     }
-    flush_count = (long) (sf_info->samplerate * (sound_latency + 0.2));
+    // 0.7 set in 2023 because output was getting truncated with lower
+    // values.  Since the truncation problem seems to be new, maybe
+    // PortAudio does not wait for CoreAudio to flush buffers and
+    // CoreAudio buffers got bigger for some reason.
+    flush_count = (long) (sf_info->samplerate * (sound_latency + 0.7));
 
     if (portaudio_error(Pa_StartStream(audio_stream),
                         "could not start audio")) {
@@ -341,7 +345,7 @@ long lookup_format(long format, long mode, long bits, long swap)
         sf_format |= (swap ? SF_ENDIAN_LITTLE : SF_ENDIAN_BIG);
 #endif
 #ifdef XL_LITTLE_ENDIAN
-        sf_format |= (swap ? SF_ENDIAN_LITTLE : SF_ENDIAN_LITTLE);
+        sf_format |= (swap ? SF_ENDIAN_BIG : SF_ENDIAN_LITTLE);
 #endif        
         break;
     case SND_HEAD_OGG: sf_format = SF_FORMAT_OGG; mode = SND_MODE_VORBIS; break;
@@ -482,11 +486,19 @@ double sound_save(LVAL snd_expr, int64_t n, unsigned char *filename,
                     sf_command(sndfile, SFC_SET_CLIPPING, NULL, SF_TRUE);
                 } else {
                     char error[240];
-                    sprintf(error, "snd_save -- %s", sf_error_number(sf_error(sndfile)));
+                    FILE *temp;
+                    if (!(temp = fopen((char *) filename, "wb"))) {
+                        snprintf(error, 240, "S-SAVE -- could not open %s",
+                                 filename);
+                    } else {
+                        fclose(temp);
+                        snprintf(error, 240, "S-SAVE -- %s",
+                                 sf_error_number(sf_error(sndfile)));
+                    }
                     xlabort(error);
                 }
             } else {
-                xlabort("snd_save -- write not permitted by -W option");
+                xlabort("S-SAVE -- write not permitted by -W option");
             }
         }
         if (play)
