@@ -1233,8 +1233,14 @@ void min_cnt(int64_t *cnt_ptr, sound_type sound, snd_susp_type susp, long cnt)
       susp->sr);
     /* if *cnt_ptr is uninitialized, just plug in c, otherwise compute min */
     if ((*cnt_ptr == UNKNOWN) || (*cnt_ptr > c)) {
-/*        nyquist_printf("min_cnt %p: new count is %d\n", susp, c);*/
-/*        if (c == 0) sound_print_tree(printing_this_sound);*/
+        /* nyquist_printf("min_cnt %p: new count is %lld\n", susp, c);
+        nyquist_printf("  current %lld, cnt %lld, sr %g, t0 %g, susp->t0 %g "
+                       "susp->sr %g\n",
+                       sound->current, cnt, sound->sr, sound->t0, susp->t0,
+                       susp->sr);
+        nyquist_printf("  before rounding %g\n", (((sound->current - cnt) /
+                sound->sr + sound->t0) - susp->t0) * susp->sr);
+        if (c == 0) sound_print_tree(printing_this_sound); */
         *cnt_ptr = c;
     }
 }
@@ -1337,7 +1343,7 @@ void set_watch(where)
         nyquist_printf("set_watch: watch_me = %p\n", watch_me);
     }
 }
-*/
+   end debugging */
 
 /*
  * additional routines
